Index: src/wp-admin/admin-ajax.php
===================================================================
--- src/wp-admin/admin-ajax.php	(revision 26185)
+++ src/wp-admin/admin-ajax.php	(working copy)
@@ -50,7 +50,7 @@
 	'oembed-cache', 'image-editor', 'delete-comment', 'delete-tag', 'delete-link',
 	'delete-meta', 'delete-post', 'trash-post', 'untrash-post', 'delete-page', 'dim-comment',
 	'add-link-category', 'add-tag', 'get-tagcloud', 'get-comments', 'replyto-comment',
-	'edit-comment', 'add-menu-item', 'add-meta', 'add-user', 'autosave', 'closed-postboxes',
+	'edit-comment', 'add-menu-item', 'add-meta', 'add-user', 'closed-postboxes',
 	'hidden-columns', 'update-welcome-panel', 'menu-get-metabox', 'wp-link-ajax',
 	'menu-locations-save', 'menu-quick-search', 'meta-box-order', 'get-permalink',
 	'sample-permalink', 'inline-save', 'inline-save-tax', 'find_posts', 'widgets-order',
Index: src/wp-admin/edit-form-advanced.php
===================================================================
--- src/wp-admin/edit-form-advanced.php	(revision 26185)
+++ src/wp-admin/edit-form-advanced.php	(working copy)
@@ -403,7 +403,6 @@
 
 echo $form_extra;
 
-wp_nonce_field( 'autosave', 'autosavenonce', false );
 wp_nonce_field( 'meta-box-order', 'meta-box-order-nonce', false );
 wp_nonce_field( 'closedpostboxes', 'closedpostboxesnonce', false );
 ?>
Index: src/wp-admin/includes/ajax-actions.php
===================================================================
--- src/wp-admin/includes/ajax-actions.php	(revision 26185)
+++ src/wp-admin/includes/ajax-actions.php	(working copy)
@@ -1084,68 +1084,6 @@
 	$x->send();
 }
 
-function wp_ajax_autosave() {
-	define( 'DOING_AUTOSAVE', true );
-
-	check_ajax_referer( 'autosave', 'autosavenonce' );
-
-	if ( ! empty( $_POST['catslist'] ) )
-		$_POST['post_category'] = explode( ',', $_POST['catslist'] );
-	if ( $_POST['post_type'] == 'page' || empty( $_POST['post_category'] ) )
-		unset( $_POST['post_category'] );
-
-	$data = '';
-	$supplemental = array();
-	$id = $revision_id = 0;
-
-	$post_id = (int) $_POST['post_id'];
-	$_POST['ID'] = $_POST['post_ID'] = $post_id;
-	$post = get_post( $post_id );
-	if ( empty( $post->ID ) || ! current_user_can( 'edit_post', $post->ID ) )
-		wp_die( __( 'You are not allowed to edit this post.' ) );
-
-	if ( 'page' == $post->post_type && ! current_user_can( 'edit_page', $post->ID ) )
-		wp_die( __( 'You are not allowed to edit this page.' ) );
-
-	if ( 'auto-draft' == $post->post_status )
-		$_POST['post_status'] = 'draft';
-
-	if ( ! empty( $_POST['autosave'] ) ) {
-		if ( ! wp_check_post_lock( $post->ID ) && get_current_user_id() == $post->post_author && ( 'auto-draft' == $post->post_status || 'draft' == $post->post_status ) ) {
-			// Drafts and auto-drafts are just overwritten by autosave for the same user if the post is not locked
-			$id = edit_post();
-		} else {
-			// Non drafts or other users drafts are not overwritten. The autosave is stored in a special post revision for each user.
-			$revision_id = wp_create_post_autosave( $post->ID );
-			if ( is_wp_error($revision_id) )
-				$id = $revision_id;
-			else
-				$id = $post->ID;
-		}
-
-		if ( ! is_wp_error($id) ) {
-			/* translators: draft saved date format, see http://php.net/date */
-			$draft_saved_date_format = __('g:i:s a');
-			/* translators: %s: date and time */
-			$data = sprintf( __('Draft saved at %s.'), date_i18n( $draft_saved_date_format ) );
-		}
-	} else {
-		if ( ! empty( $_POST['auto_draft'] ) )
-			$id = 0; // This tells us it didn't actually save
-		else
-			$id = $post->ID;
-	}
-
-	// @todo Consider exposing any errors, rather than having 'Saving draft...'
-	$x = new WP_Ajax_Response( array(
-		'what' => 'autosave',
-		'id' => $id,
-		'data' => $data,
-		'supplemental' => $supplemental
-	) );
-	$x->send();
-}
-
 function wp_ajax_closed_postboxes() {
 	check_ajax_referer( 'closedpostboxes', 'closedpostboxesnonce' );
 	$closed = isset( $_POST['closed'] ) ? explode( ',', $_POST['closed']) : array();
Index: src/wp-admin/includes/misc.php
===================================================================
--- src/wp-admin/includes/misc.php	(revision 26185)
+++ src/wp-admin/includes/misc.php	(working copy)
@@ -724,7 +724,6 @@
 		if ( 2 === wp_verify_nonce( $received['post_nonce'], 'update-post_' . $post_id ) ) {
 			$response['wp-refresh-post-nonces'] = array(
 				'replace' => array(
-					'autosavenonce' => wp_create_nonce('autosave'),
 					'getpermalinknonce' => wp_create_nonce('getpermalink'),
 					'samplepermalinknonce' => wp_create_nonce('samplepermalink'),
 					'closedpostboxesnonce' => wp_create_nonce('closedpostboxes'),
@@ -739,3 +738,29 @@
 	return $response;
 }
 add_filter( 'heartbeat_received', 'wp_refresh_post_nonces', 10, 3 );
+
+/**
+ * Autosave with heartbeat
+ *
+ * @since 3.8
+ */
+function heartbeat_autosave( $response, $data ) {
+	if ( ! empty( $data['wp_autosave'] ) ) {
+		$saved = wp_autosave( $data['wp_autosave'] );
+
+		if ( is_wp_error( $saved ) ) {
+			$response['wp_autosave'] = array( 'success' => false, 'message' => $saved->get_error_message() );
+		} elseif ( empty( $saved ) ) {
+			$response['wp_autosave'] = array( 'success' => false, 'message' => __( 'Error while saving.' ) );
+		} else {
+			/* translators: draft saved date format, see http://php.net/date */
+			$draft_saved_date_format = __( 'g:i:s a' );
+			/* translators: %s: date and time */
+			$response['wp_autosave'] = array( 'success' => true, 'message' => sprintf( __( 'Draft saved at %s.' ), date_i18n( $draft_saved_date_format ) ) );
+		}
+	}
+
+	return $response;
+}
+// Run later as we have to set DOING_AUTOSAVE for back-compat
+add_filter( 'heartbeat_received', 'heartbeat_autosave', 50, 2 );
Index: src/wp-admin/includes/post.php
===================================================================
--- src/wp-admin/includes/post.php	(revision 26185)
+++ src/wp-admin/includes/post.php	(working copy)
@@ -79,9 +79,14 @@
 		}
 	}
 
-	if ( ! empty( $post_data['post_status'] ) )
+	if ( ! empty( $post_data['post_status'] ) ) {
 		$post_data['post_status'] = sanitize_key( $post_data['post_status'] );
 
+		// No longer an auto-draft
+		if ( 'auto-draft' == $post_data['post_status'] )
+			$post_data['post_status'] = 'draft';
+	}
+
 	// What to do based on which button they pressed
 	if ( isset($post_data['saveasdraft']) && '' != $post_data['saveasdraft'] )
 		$post_data['post_status'] = 'draft';
@@ -190,9 +195,6 @@
 	$post_data = _wp_translate_postdata( true, $post_data );
 	if ( is_wp_error($post_data) )
 		wp_die( $post_data->get_error_message() );
-	if ( ( empty( $post_data['action'] ) || 'autosave' != $post_data['action'] ) && 'auto-draft' == $post_data['post_status'] ) {
-		$post_data['post_status'] = 'draft';
-	}
 
 	if ( isset($post_data['visibility']) ) {
 		switch ( $post_data['visibility'] ) {
@@ -1335,22 +1337,30 @@
  * @uses _wp_translate_postdata()
  * @uses _wp_post_revision_fields()
  *
- * @return unknown
+ * @param mixed $post_data Associative array containing the post data or int post ID.
+ * @return mixed The autosave revision ID. WP_Error or 0 on error.
  */
-function wp_create_post_autosave( $post_id ) {
-	$translated = _wp_translate_postdata( true );
-	if ( is_wp_error( $translated ) )
-		return $translated;
+function wp_create_post_autosave( $post_data ) {
+	if ( is_numeric( $post_data ) ) {
+		$post_id = $post_data;
+		$post_data = &$_POST;
+	} else {
+		$post_id = (int) $post_data['post_ID'];
+	}
 
+	$post_data = _wp_translate_postdata( true, $post_data );
+	if ( is_wp_error( $post_data ) )
+		return $post_data;
+
 	$post_author = get_current_user_id();
 
 	// Store one autosave per author. If there is already an autosave, overwrite it.
 	if ( $old_autosave = wp_get_post_autosave( $post_id, $post_author ) ) {
-		$new_autosave = _wp_post_revision_fields( $_POST, true );
+		$new_autosave = _wp_post_revision_fields( $post_data, true );
 		$new_autosave['ID'] = $old_autosave->ID;
 		$new_autosave['post_author'] = $post_author;
 
-		// If the new autosave is the same content as the post, delete the old autosave.
+		// If the new autosave has the same content as the post, delete the autosave.
 		$post = get_post( $post_id );
 		$autosave_is_different = false;
 		foreach ( array_keys( _wp_post_revision_fields() ) as $field ) {
@@ -1362,7 +1372,7 @@
 
 		if ( ! $autosave_is_different ) {
 			wp_delete_post_revision( $old_autosave->ID );
-			return;
+			return 0;
 		}
 
 		return wp_update_post( $new_autosave );
@@ -1369,7 +1379,7 @@
 	}
 
 	// _wp_put_post_revision() expects unescaped.
-	$post_data = wp_unslash( $_POST );
+	$post_data = wp_unslash( $post_data );
 
 	// Otherwise create the new autosave as a special post revision
 	return _wp_put_post_revision( $post_data, true );
@@ -1395,58 +1405,82 @@
 function post_preview() {
 
 	$post_ID = (int) $_POST['post_ID'];
-	$status = get_post_status( $post_ID );
-	if ( 'auto-draft' == $status )
-		wp_die( __('Preview not available. Please save as a draft first.') );
+	$_POST['ID'] = $post_ID;
 
-	if ( isset($_POST['catslist']) )
-		$_POST['post_category'] = explode(",", $_POST['catslist']);
+	if ( ! $post = get_post( $post_ID ) )
+		wp_die( __('You attempted to preview a non existing item.') );
 
-	if ( isset($_POST['tags_input']) )
-		$_POST['tags_input'] = explode(",", $_POST['tags_input']);
+	if ( ! current_user_can( 'edit_post', $post->ID ) )
+		wp_die( __('You are not allowed to preview this item.') );
 
-	if ( $_POST['post_type'] == 'page' || empty($_POST['post_category']) )
-		unset($_POST['post_category']);
+	$is_autosave = false;
 
-	$_POST['ID'] = $post_ID;
-	$post = get_post($post_ID);
-
-	if ( 'page' == $post->post_type ) {
-		if ( ! current_user_can('edit_page', $post_ID) )
-			wp_die( __('You are not allowed to edit this page.') );
+	if ( ! wp_check_post_lock( $post->ID ) && get_current_user_id() == $post->post_author && ( 'draft' == $post->post_status || 'auto-draft' == $post->post_status ) ) {
+		$saved_post_id = edit_post();
 	} else {
-		if ( ! current_user_can('edit_post', $post_ID) )
-			wp_die( __('You are not allowed to edit this post.') );
-	}
+		$is_autosave = true;
 
-	$user_id = get_current_user_id();
-	$locked = wp_check_post_lock( $post->ID );
-	if ( ! $locked && 'draft' == $post->post_status && $user_id == $post->post_author ) {
-		$id = edit_post();
-	} else { // Non drafts are not overwritten. The autosave is stored in a special post revision.
-		$id = wp_create_post_autosave( $post->ID );
-		if ( ! is_wp_error($id) )
-			$id = $post->ID;
+		if ( 'auto-draft' == $_POST['post_status'] )
+			$_POST['post_status'] = 'draft';
+
+		$saved_post_id = wp_create_post_autosave( $post->ID );
 	}
 
-	if ( is_wp_error($id) )
-		wp_die( $id->get_error_message() );
+	if ( is_wp_error( $saved_post_id ) )
+		wp_die( $saved_post_id->get_error_message() );
 
-	if ( ! $locked && $_POST['post_status'] == 'draft' && $user_id == $post->post_author ) {
-		$url = add_query_arg( 'preview', 'true', get_permalink($id) );
-	} else {
-		$nonce = wp_create_nonce('post_preview_' . $id);
-		$args = array(
-			'preview' => 'true',
-			'preview_id' => $id,
-			'preview_nonce' => $nonce,
-		);
+	$query_args = array( 'preview' => 'true' );
 
+	if ( $is_autosave && $saved_post_id ) {
+		$query_args['preview_id'] = $post->ID;
+		$query_args['preview_nonce'] = wp_create_nonce( 'post_preview_' . $post->ID );
+
 		if ( isset( $_POST['post_format'] ) )
-			$args['post_format'] = empty( $_POST['post_format'] ) ? 'standard' : sanitize_key( $_POST['post_format'] );
-
-		$url = add_query_arg( $args, get_permalink($id) );
+			$query_args['post_format'] = empty( $_POST['post_format'] ) ? 'standard' : sanitize_key( $_POST['post_format'] );
 	}
 
+	$url = add_query_arg( $query_args, get_permalink( $post->ID ) );
 	return apply_filters( 'preview_post_link', $url );
 }
+
+/**
+ * Save a post submitted with XHR
+ *
+ * Intended for use with heartbeat and autosave.js
+ *
+ * @since 3.8
+ *
+ * @param $post_data Associative array of the submitted post data.
+ * @return mixed The value 0 or WP_Error on failure. The saved post ID on success.
+ *               Te ID can be the draft post_id or the autosave revision post_id.
+ */
+function wp_autosave( $post_data ) {
+	// Back-compat
+	if ( ! defined( 'DOING_AUTOSAVE' ) )
+		define( 'DOING_AUTOSAVE', true );
+
+	$post_id = (int) $post_data['post_id'];
+	$post_data['ID'] = $post_data['post_ID'] = $post_id;
+
+	if ( false === wp_verify_nonce( $post_data['_wpnonce'], 'update-post_' . $post_id ) )
+		return new WP_Error( 'invalid_nonce', __('ERROR: invalid post data.') );
+
+	$post = get_post( $post_id );
+
+	if ( ! current_user_can( 'edit_post', $post->ID ) )
+		return new WP_Error( 'edit_post', __('You are not allowed to edit this item.') );
+
+	if ( 'auto-draft' == $post->post_status )
+		$post_data['post_status'] = 'draft';
+
+	if ( $post_data['post_type'] != 'page' && ! empty( $post_data['catslist'] ) )
+		$post_data['post_category'] = explode( ',', $post_data['catslist'] );
+
+	if ( ! wp_check_post_lock( $post->ID ) && get_current_user_id() == $post->post_author && ( 'auto-draft' == $post->post_status || 'draft' == $post->post_status ) ) {
+		// Drafts and auto-drafts are just overwritten by autosave for the same user if the post is not locked
+		return edit_post( $post_data );
+	} else {
+		// Non drafts or other users drafts are not overwritten. The autosave is stored in a special post revision for each user.
+		return wp_create_post_autosave( $post_data );
+	}
+}
Index: src/wp-admin/js/post.js
===================================================================
--- src/wp-admin/js/post.js	(revision 26185)
+++ src/wp-admin/js/post.js	(working copy)
@@ -1,4 +1,6 @@
-var tagBox, commentsBox, editPermalink, makeSlugeditClickable, WPSetThumbnailHTML, WPSetThumbnailID, WPRemoveThumbnail, wptitlehint;
+var tagBox, commentsBox, WPSetThumbnailHTML, WPSetThumbnailID, WPRemoveThumbnail, wptitlehint;
+// Prevent fatal errors (back-compat)
+makeSlugeditClickable = editPermalink = function(){};
 
 // return an array with any duplicate, whitespace or values removed
 function array_unique_noempty(a) {
@@ -265,10 +267,9 @@
 		send['lock'] = lock;
 
 	data['wp-refresh-post-lock'] = send;
-});
 
-// Post locks: update the lock string or show the dialog if somebody has taken over editing
-$(document).on( 'heartbeat-tick.refresh-lock', function( e, data ) {
+}).on( 'heartbeat-tick.refresh-lock', function( e, data ) {
+	// Post locks: update the lock string or show the dialog if somebody has taken over editing
 	var received, wrap, avatar;
 
 	if ( data['wp-refresh-post-lock'] ) {
@@ -279,19 +280,16 @@
 			wrap = $('#post-lock-dialog');
 
 			if ( wrap.length && ! wrap.is(':visible') ) {
-				if ( typeof autosave == 'function' ) {
-					$(document).on('autosave-disable-buttons.post-lock', function() {
-						wrap.addClass('saving');
-					}).on('autosave-enable-buttons.post-lock', function() {
+				if ( typeof wp != 'undefined' && wp.autosave ) {
+					// Save the latest changes and disable
+					$(document).one( 'heartbeat-tick', function() {
+						wp.autosave.server.disable();
 						wrap.removeClass('saving').addClass('saved');
-						window.onbeforeunload = null;
+						$(window).off( 'beforeunload.autosave' );
 					});
 
-					// Save the latest changes and disable
-					if ( ! autosave() )
-						window.onbeforeunload = null;
-
-					autosave = function(){};
+					wrap.addClass('saving');
+					wp.autosave.server.triggerSave();
 				}
 
 				if ( received.lock_error.avatar_src ) {
@@ -306,6 +304,22 @@
 			$('#active_post_lock').val( received.new_lock );
 		}
 	}
+}).on( 'after-autosave.update-post-slug', function() {
+	// create slug area only if not already there
+	if ( ! $('#edit-slug-box > *').length ) {
+		$.post( ajaxurl, {
+				action: 'sample-permalink',
+				post_id: $('#post_ID').val(),
+				new_title: typeof fullscreen != 'undefined' && fullscreen.settings.visible ? $('#wp-fullscreen-title').val() : $('#title').val(),
+				samplepermalinknonce: $('#samplepermalinknonce').val()
+			},
+			function( data ) {
+				if ( data != '-1' ) {
+					$('#edit-slug-box').html(data);
+				}
+			}
+		);
+	}
 });
 
 }(jQuery));
@@ -351,7 +365,8 @@
 }(jQuery));
 
 jQuery(document).ready( function($) {
-	var stamp, visibility, sticky = '', last = 0, co = $('#content');
+	var stamp, visibility, sticky = '', last = 0, co = $('#content'), $editSlugWrap = $('#edit-slug-box'),
+		postId = $('#post_ID').val() || 0, $submitpost = $('#submitpost'), releaseLock = true;
 
 	postboxes.add_postbox_toggles(pagenow);
 
@@ -376,6 +391,128 @@
 		wp.heartbeat.setInterval( 15 );
 	}
 
+	// The form is being submitted by the user
+	$submitpost.find( 'input[type="submit"], a.submitdelete' ).on( 'click.autosave', function( event ) {
+		var $button = $(this);
+
+		if ( $button.prop('disabled') ) {
+			event.preventDefault();
+			return;
+		}
+
+		if ( $button.hasClass('submitdelete') ) {
+			return;
+		}
+
+		// The form submission can be blocked from JS or by using HTML 5.0 validation on some fields.
+		// Run this only on an actual 'submit'.
+		$('form#post').off( 'submit.edit-post' ).on( 'submit.edit-post', function( event ) {
+			if ( event.isDefaultPrevented() )
+				return;
+
+			wp.autosave.server.disable();
+			releaseLock = false;
+			$(window).off( 'beforeunload.edit-post' );
+
+			$submitpost.find('a.submitdelete, #post-preview').addClass('button-disabled');
+
+			if ( $button.attr('id') == 'publish' )
+				$submitpost.find('#major-publishing-actions .spinner').show();
+			else
+				$submitpost.find('#minor-publishing .spinner').show();
+		});
+	});
+
+	// Submit the form saving a draft or an autosave, and show a preview in a new tab
+	$('#post-preview').on( 'click.post-preview', function( event ) {
+		var $this = $(this),
+			$form = $('form#post'),
+			$previewField = $('input#wp-preview'),
+			target = $this.attr('target') || 'wp-preview';
+		
+		event.preventDefault();
+
+		if ( $this.prop('disabled') ) {
+			return;
+		}
+
+		wp.autosave.server.blockSave();
+		$previewField.val('dopreview');
+		$form.attr( 'target', target ).submit().attr( 'target', '' );
+
+		// Workaround for WebKit bug preventing a form submitting twice to the same action.
+		// https://bugs.webkit.org/show_bug.cgi?id=28633
+		var ua = navigator.userAgent.toLowerCase();
+		if ( ua.indexOf('safari') != -1 && ua.indexOf('chrome') == -1 ) {
+			$form.attr( 'action', function( index, value ) {
+				return value + '?t=' + new Date().getTime();
+			});
+		}
+
+		$previewField.val('');
+	});
+
+	// This code is meant to allow tabbing from Title to Post content.
+	$('#title').on( 'keydown.editor-focus', function( event ) {
+		var editor;
+
+		if ( event.which == 9 && ! event.ctrlKey && ! event.altKey && ! event.shiftKey ) {
+			editor = typeof tinymce != 'undefined' && tinymce.get('content');
+
+			if ( editor && ! editor.isHidden() ) {
+				$(this).one( 'keyup', function() {
+					$('#content_tbl td.mceToolbar > a').focus();
+				});
+			} else {
+				$('#content').focus();
+			}
+
+			event.preventDefault();
+		}
+	});
+
+	$(window).on( 'beforeunload.edit-post', function() {
+		var editor = typeof tinymce != 'undefined' && tinymce.activeEditor, compareString;
+
+		if ( editor && ! editor.isHidden() ) {
+			if ( editor.isDirty() )
+				return autosaveL10n.saveAlert;
+		} else {
+			if ( fullscreen && fullscreen.settings.visible ) {
+				compareString = wp.autosave.getCompareString({
+					post_title: $('#wp-fullscreen-title').val() || '',
+					content: $('#wp_mce_fullscreen').val() || '',
+					excerpt: $('#excerpt').val() || ''
+				});
+			} else {
+				compareString = wp.autosave.getCompareString();
+			}
+
+			if ( compareString !== wp.autosave.server.autosaveLast )
+				return autosaveL10n.saveAlert;
+		}
+	}).on( 'unload.edit-post', function( event ) {
+		if ( ! releaseLock )
+			return;
+
+		// Unload is triggered (by hand) on removing the Thickbox iframe.
+		// Make sure we process only the main document unload.
+		if ( event.target && event.target.nodeName != '#document' )
+			return;
+
+		$.ajax({
+			type: 'POST',
+			url: ajaxurl,
+			async: false,
+			data: {
+				action: 'wp-remove-post-lock',
+				_wpnonce: $('#_wpnonce').val(),
+				post_ID: $('#post_ID').val(),
+				active_post_lock: $('#active_post_lock').val()
+			}
+		});
+	});
+
 	// multi-taxonomies
 	if ( $('#tagsdiv-post_tag').length ) {
 		tagBox.init();
@@ -685,9 +822,11 @@
 	} // end submitdiv
 
 	// permalink
-	if ( $('#edit-slug-box').length ) {
-		editPermalink = function(post_id) {
-			var i, c = 0, e = $('#editable-post-name'), revert_e = e.html(), real_slug = $('#post_name'), revert_slug = real_slug.val(), b = $('#edit-slug-buttons'), revert_b = b.html(), full = $('#editable-post-name-full').html();
+	if ( $editSlugWrap.length ) {
+		function editPermalink() {
+			var i, c = 0, e = $('#editable-post-name'), revert_e = e.html(), real_slug = $('#post_name'),
+				revert_slug = real_slug.val(), b = $('#edit-slug-buttons'), revert_b = b.html(),
+				full = $('#editable-post-name-full').html();
 
 			$('#view-post-btn').hide();
 			b.html('<a href="#" class="save button button-small">'+postL10n.ok+'</a> <a class="cancel" href="#">'+postL10n.cancel+'</a>');
@@ -694,11 +833,11 @@
 			b.children('.save').click(function() {
 				var new_slug = e.children('input').val();
 				if ( new_slug == $('#editable-post-name-full').text() ) {
-					return $('.cancel', '#edit-slug-buttons').click();
+					return $('#edit-slug-buttons .cancel').click();
 				}
 				$.post(ajaxurl, {
 					action: 'sample-permalink',
-					post_id: post_id,
+					post_id: postId,
 					new_slug: new_slug,
 					new_title: $('#title').val(),
 					samplepermalinknonce: $('#samplepermalinknonce').val()
@@ -712,13 +851,12 @@
 					}
 					b.html(revert_b);
 					real_slug.val(new_slug);
-					makeSlugeditClickable();
 					$('#view-post-btn').show();
 				});
 				return false;
 			});
 
-			$('.cancel', '#edit-slug-buttons').click(function() {
+			$('#edit-slug-buttons .cancel').click(function() {
 				$('#view-post-btn').show();
 				e.html(revert_e);
 				b.html(revert_b);
@@ -748,12 +886,13 @@
 			}).focus();
 		}
 
-		makeSlugeditClickable = function() {
-			$('#editable-post-name').click(function() {
-				$('#edit-slug-buttons').children('.edit-slug').click();
-			});
-		}
-		makeSlugeditClickable();
+		$editSlugWrap.on( 'click', function( event ) {
+			var $target = $( event.target );
+
+			if ( $target.is('#editable-post-name') || $target.hasClass('edit-slug') ) {
+				editPermalink();
+			}
+		});
 	}
 
 	// word count
Index: src/wp-admin/post.php
===================================================================
--- src/wp-admin/post.php	(revision 26185)
+++ src/wp-admin/post.php	(working copy)
@@ -304,7 +304,7 @@
 	break;
 
 case 'preview':
-	check_admin_referer( 'autosave', 'autosavenonce' );
+	check_admin_referer( 'update-post_' . $post_id );
 
 	$url = post_preview();
 
Index: src/wp-includes/js/autosave.js
===================================================================
--- src/wp-includes/js/autosave.js	(revision 26185)
+++ src/wp-includes/js/autosave.js	(working copy)
@@ -1,400 +1,106 @@
-var autosave, autosaveLast = '', autosavePeriodical, autosaveDelayPreview = false, notSaved = true, blockSave = false, fullscreen, autosaveLockRelease = true;
+/* global tinymce, wpCookies, autosaveL10n, fullscreen, switchEditors */
+window.wp = window.wp || {};
 
-jQuery(document).ready( function($) {
+( function( $, window ) {
+	var _disabled, _blockSave, _blockSaveTimer,
+		$document = $(document),
+		nextRun = 0;
 
-	if ( $('#wp-content-wrap').hasClass('tmce-active') && typeof switchEditors != 'undefined' ) {
-		autosaveLast = wp.autosave.getCompareString({
-			post_title : $('#title').val() || '',
-			content : switchEditors.pre_wpautop( $('#content').val() ) || '',
-			excerpt : $('#excerpt').val() || ''
-		});
-	} else {
-		autosaveLast = wp.autosave.getCompareString();
-	}
+/**
+ * Returns the data saved in both local and remote autosave
+ *
+ * @return mixed Object containing the post data or false
+ */
+wp.autosave = {
+	getPostData: function( type ) {
+		var post_name, parent_id, data,
+			cats = [],
+			editor = typeof tinymce !== 'undefined' && tinymce.get('content');
 
-	autosavePeriodical = $.schedule({time: autosaveL10n.autosaveInterval * 1000, func: function() { autosave(); }, repeat: true, protect: true});
-
-	//Disable autosave after the form has been submitted
-	$("#post").submit(function() {
-		$.cancel(autosavePeriodical);
-		autosaveLockRelease = false;
-	});
-
-	$('input[type="submit"], a.submitdelete', '#submitpost').click(function(){
-		blockSave = true;
-		window.onbeforeunload = null;
-		$(':button, :submit', '#submitpost').each(function(){
-			var t = $(this);
-			if ( t.hasClass('button-primary') )
-				t.addClass('button-primary-disabled');
-			else
-				t.addClass('button-disabled');
-		});
-		if ( $(this).attr('id') == 'publish' )
-			$('#major-publishing-actions .spinner').show();
-		else
-			$('#minor-publishing .spinner').show();
-	});
-
-	window.onbeforeunload = function(){
-		var editor = typeof(tinymce) != 'undefined' ? tinymce.activeEditor : false, compareString;
-
 		if ( editor && ! editor.isHidden() ) {
-			if ( editor.isDirty() )
-				return autosaveL10n.saveAlert;
-		} else {
-			if ( fullscreen && fullscreen.settings.visible ) {
-				compareString = wp.autosave.getCompareString({
-					post_title: $('#wp-fullscreen-title').val() || '',
-					content: $('#wp_mce_fullscreen').val() || '',
-					excerpt: $('#excerpt').val() || ''
-				});
+			// Don't run while the tinymce spellcheck is on. It resets all found words.
+			if ( editor.plugins.spellchecker && editor.plugins.spellchecker.active ) {
+				return false;
 			} else {
-				compareString = wp.autosave.getCompareString();
+				tinymce.triggerSave();
 			}
-
-			if ( compareString != autosaveLast )
-				return autosaveL10n.saveAlert;
 		}
-	};
 
-	$(window).unload( function(e) {
-		if ( ! autosaveLockRelease )
-			return;
+		data = {
+			post_id: $( '#post_ID' ).val() || 0,
+			post_type: $( '#post_type' ).val() || '',
+			post_author: $( '#post_author' ).val() || '',
+			excerpt: $( '#excerpt' ).val() || ''
+		};
 
-		// unload fires (twice) on removing the Thickbox iframe. Make sure we process only the main document unload.
-		if ( e.target && e.target.nodeName != '#document' )
-			return;
-
-		$.ajax({
-			type: 'POST',
-			url: ajaxurl,
-			async: false,
-			data: {
-				action: 'wp-remove-post-lock',
-				_wpnonce: $('#_wpnonce').val(),
-				post_ID: $('#post_ID').val(),
-				active_post_lock: $('#active_post_lock').val()
-			}
-		});
-	} );
-
-	// preview
-	$('#post-preview').click(function(){
-		if ( $('#auto_draft').val() == '1' && notSaved ) {
-			autosaveDelayPreview = true;
-			autosave();
-			return false;
+		if ( typeof fullscreen !== 'undefined' && fullscreen.settings.visible ) {
+			data.post_title = $( '#wp-fullscreen-title' ).val() || '';
+			data.content = $( '#wp_mce_fullscreen' ).val() || '';
+		} else {
+			data.post_title = $( '#title' ).val() || '';
+			data.content = $( '#content' ).val() || '';
 		}
-		doPreview();
-		return false;
-	});
 
-	doPreview = function() {
-		$('input#wp-preview').val('dopreview');
-		$('form#post').attr('target', 'wp-preview').submit().attr('target', '');
-
-		/*
-		 * Workaround for WebKit bug preventing a form submitting twice to the same action.
-		 * https://bugs.webkit.org/show_bug.cgi?id=28633
-		 */
-		var ua = navigator.userAgent.toLowerCase();
-		if ( ua.indexOf('safari') != -1 && ua.indexOf('chrome') == -1 ) {
-			$('form#post').attr('action', function(index, value) {
-				return value + '?t=' + new Date().getTime();
-			});
+		if ( type === 'local' ) {
+			return data;
 		}
 
-		$('input#wp-preview').val('');
-	}
+		$( 'input[id^="in-category-"]:checked' ).each( function() {
+			cats.push( this.value );
+		});
+		data.catslist = cats.join(',');
 
-	// This code is meant to allow tabbing from Title to Post content.
-	$('#title').on('keydown.editor-focus', function(e) {
-		var ed;
-
-		if ( e.which != 9 )
-			return;
-
-		if ( !e.ctrlKey && !e.altKey && !e.shiftKey ) {
-			if ( typeof(tinymce) != 'undefined' )
-				ed = tinymce.get('content');
-
-			if ( ed && !ed.isHidden() ) {
-				$(this).one('keyup', function(e){
-					$('#content_tbl td.mceToolbar > a').focus();
-				});
-			} else {
-				$('#content').focus();
-			}
-
-			e.preventDefault();
+		if ( post_name = $( '#post_name' ).val() ) {
+			data.post_name = post_name;
 		}
-	});
 
-	// autosave new posts after a title is typed but not if Publish or Save Draft is clicked
-	if ( '1' == $('#auto_draft').val() ) {
-		$('#title').blur( function() {
-			if ( !this.value || $('#auto_draft').val() != '1' )
-				return;
-			delayed_autosave();
-		});
-	}
-
-	// When connection is lost, keep user from submitting changes.
-	$(document).on('heartbeat-connection-lost.autosave', function( e, error, status ) {
-		if ( 'timeout' === error || 503 == status ) {
-			var notice = $('#lost-connection-notice');
-			if ( ! wp.autosave.local.hasStorage ) {
-				notice.find('.hide-if-no-sessionstorage').hide();
-			}
-			notice.show();
-			autosave_disable_buttons();
+		if ( parent_id = $( '#parent_id' ).val() ) {
+			data.parent_id = parent_id;
 		}
-	}).on('heartbeat-connection-restored.autosave', function() {
-		$('#lost-connection-notice').hide();
-		autosave_enable_buttons();
-	});
-});
 
-function autosave_parse_response( response ) {
-	var res = wpAjax.parseAjaxResponse(response, 'autosave'), post_id, sup;
-
-	if ( res && res.responses && res.responses.length ) {
-		if ( res.responses[0].supplemental ) {
-			sup = res.responses[0].supplemental;
-
-			jQuery.each( sup, function( selector, value ) {
-				if ( selector.match(/^replace-/) )
-					jQuery( '#' + selector.replace('replace-', '') ).val( value );
-			});
+		if ( $( '#comment_status' ).prop( 'checked' ) ) {
+			data.comment_status = 'open';
 		}
 
-		// if no errors: add slug UI and update autosave-message
-		if ( !res.errors ) {
-			if ( post_id = parseInt( res.responses[0].id, 10 ) )
-				autosave_update_slug( post_id );
-
-			if ( res.responses[0].data ) // update autosave message
-				jQuery('.autosave-message').text( res.responses[0].data );
+		if ( $( '#ping_status' ).prop( 'checked' ) ) {
+			data.ping_status = 'open';
 		}
-	}
 
-	return res;
-}
-
-// called when autosaving pre-existing post
-function autosave_saved(response) {
-	blockSave = false;
-	autosave_parse_response(response); // parse the ajax response
-	autosave_enable_buttons(); // re-enable disabled form buttons
-}
-
-// called when autosaving new post
-function autosave_saved_new(response) {
-	blockSave = false;
-	var res = autosave_parse_response(response), post_id;
-
-	if ( res && res.responses.length && !res.errors ) {
-		// An ID is sent only for real auto-saves, not for autosave=0 "keepalive" saves
-		post_id = parseInt( res.responses[0].id, 10 );
-
-		if ( post_id ) {
-			notSaved = false;
-			jQuery('#auto_draft').val('0'); // No longer an auto-draft
+		if ( $( '#auto_draft' ).val() === '1' ) {
+			data.auto_draft = '1';
 		}
 
-		autosave_enable_buttons();
+		return data;
+	},
 
-		if ( autosaveDelayPreview ) {
-			autosaveDelayPreview = false;
-			doPreview();
+	// Concatenate title, content and excerpt. Used to track changes when auto-saving.
+	getCompareString: function( post_data ) {
+		if ( typeof post_data === 'object' ) {
+			return ( post_data.post_title || '' ) + '::' + ( post_data.content || '' ) + '::' + ( post_data.excerpt || '' );
 		}
-	} else {
-		autosave_enable_buttons(); // re-enable disabled form buttons
-	}
-}
 
-function autosave_update_slug(post_id) {
-	// create slug area only if not already there
-	if ( 'undefined' != makeSlugeditClickable && jQuery.isFunction(makeSlugeditClickable) && !jQuery('#edit-slug-box > *').size() ) {
-		jQuery.post( ajaxurl, {
-				action: 'sample-permalink',
-				post_id: post_id,
-				new_title: fullscreen && fullscreen.settings.visible ? jQuery('#wp-fullscreen-title').val() : jQuery('#title').val(),
-				samplepermalinknonce: jQuery('#samplepermalinknonce').val()
-			},
-			function(data) {
-				if ( data !== '-1' ) {
-					var box = jQuery('#edit-slug-box');
-					box.html(data);
-					if (box.hasClass('hidden')) {
-						box.fadeIn('fast', function () {
-							box.removeClass('hidden');
-						});
-					}
-					makeSlugeditClickable();
-				}
-			}
-		);
-	}
-}
+		return ( $('#title').val() || '' ) + '::' + ( $('#content').val() || '' ) + '::' + ( $('#excerpt').val() || '' );
+	},
 
-function autosave_loading() {
-	jQuery('.autosave-message').html(autosaveL10n.savingText);
-}
+	enableButtons: function() {
+		$document.trigger( 'autosave-enable-buttons' );
 
-function autosave_enable_buttons() {
-	jQuery(document).trigger('autosave-enable-buttons');
-	if ( ! wp.heartbeat || ! wp.heartbeat.hasConnectionError() ) {
-		// delay that a bit to avoid some rare collisions while the DOM is being updated.
-		setTimeout(function(){
-			var parent = jQuery('#submitpost');
-			parent.find(':button, :submit').removeAttr('disabled');
-			parent.find('.spinner').hide();
-		}, 500);
-	}
-}
+		if ( ! wp.heartbeat || ! wp.heartbeat.hasConnectionError() ) {
+			$( '#submitpost' ).find( ':button, :submit, a.submitdelete, #post-preview' ).prop( 'disabled', false ).removeClass( 'button-disabled' );
+		}
+	},
 
-function autosave_disable_buttons() {
-	jQuery(document).trigger('autosave-disable-buttons');
-	jQuery('#submitpost').find(':button, :submit').prop('disabled', true);
-	// Re-enable 5 sec later. Just gives autosave a head start to avoid collisions.
-	setTimeout( autosave_enable_buttons, 5000 );
-}
-
-function delayed_autosave() {
-	setTimeout(function(){
-		if ( blockSave )
-			return;
-		autosave();
-	}, 200);
-}
-
-autosave = function() {
-	var post_data = wp.autosave.getPostData(),
-		compareString,
-		successCallback;
-
-	blockSave = true;
-
-	// post_data.content cannot be retrieved at the moment
-	if ( ! post_data.autosave )
-		return false;
-
-	// No autosave while thickbox is open (media buttons)
-	if ( jQuery("#TB_window").css('display') == 'block' )
-		return false;
-
-	compareString = wp.autosave.getCompareString( post_data );
-
-	// Nothing to save or no change.
-	if ( compareString == autosaveLast )
-		return false;
-
-	autosaveLast = compareString;
-	jQuery(document).triggerHandler('wpcountwords', [ post_data["content"] ]);
-
-	// Disable buttons until we know the save completed.
-	autosave_disable_buttons();
-
-	if ( post_data["auto_draft"] == '1' ) {
-		successCallback = autosave_saved_new; // new post
-	} else {
-		successCallback = autosave_saved; // pre-existing post
+	disableButtons: function() {
+		$document.trigger('autosave-disable-buttons');
+		$( '#submitpost' ).find( ':button, :submit, a.submitdelete, #post-preview' ).prop( 'disabled', true ).addClass( 'button-disabled' );
+		// Re-enable 5 sec later. Just gives autosave a head start to avoid collisions.
+		setTimeout( wp.autosave.enableButtons, 5000 );
 	}
+};
 
-	jQuery.ajax({
-		data: post_data,
-		beforeSend: autosave_loading,
-		type: "POST",
-		url: ajaxurl,
-		success: successCallback
-	});
-
-	return true;
-}
-
 // Autosave in localStorage
 // set as simple object/mixin for now
-window.wp = window.wp || {};
-wp.autosave = wp.autosave || {};
-
-(function($){
-// Returns the data for saving in both localStorage and autosaves to the server
-wp.autosave.getPostData = function() {
-	var ed = typeof tinymce != 'undefined' ? tinymce.activeEditor : null, post_name, parent_id, cats = [],
-		data = {
-			action: 'autosave',
-			autosave: true,
-			post_id: $('#post_ID').val() || 0,
-			autosavenonce: $('#autosavenonce').val() || '',
-			post_type: $('#post_type').val() || '',
-			post_author: $('#post_author').val() || '',
-			excerpt: $('#excerpt').val() || ''
-		};
-
-	if ( ed && !ed.isHidden() ) {
-		// Don't run while the tinymce spellcheck is on. It resets all found words.
-		if ( ed.plugins.spellchecker && ed.plugins.spellchecker.active ) {
-			data.autosave = false;
-			return data;
-		} else {
-			if ( 'mce_fullscreen' == ed.id )
-				tinymce.get('content').setContent(ed.getContent({format : 'raw'}), {format : 'raw'});
-
-			tinymce.triggerSave();
-		}
-	}
-
-	if ( typeof fullscreen != 'undefined' && fullscreen.settings.visible ) {
-		data['post_title'] = $('#wp-fullscreen-title').val() || '';
-		data['content'] = $('#wp_mce_fullscreen').val() || '';
-	} else {
-		data['post_title'] = $('#title').val() || '';
-		data['content'] = $('#content').val() || '';
-	}
-
-	/*
-	// We haven't been saving tags with autosave since 2.8... Start again?
-	$('.the-tags').each( function() {
-		data[this.name] = this.value;
-	});
-	*/
-
-	$('input[id^="in-category-"]:checked').each( function() {
-		cats.push(this.value);
-	});
-	data['catslist'] = cats.join(',');
-
-	if ( post_name = $('#post_name').val() )
-		data['post_name'] = post_name;
-
-	if ( parent_id = $('#parent_id').val() )
-		data['parent_id'] = parent_id;
-
-	if ( $('#comment_status').prop('checked') )
-		data['comment_status'] = 'open';
-
-	if ( $('#ping_status').prop('checked') )
-		data['ping_status'] = 'open';
-
-	if ( $('#auto_draft').val() == '1' )
-		data['auto_draft'] = '1';
-
-	return data;
-};
-
-// Concatenate title, content and excerpt. Used to track changes when auto-saving.
-wp.autosave.getCompareString = function( post_data ) {
-	if ( typeof post_data === 'object' ) {
-		return ( post_data.post_title || '' ) + '::' + ( post_data.content || '' ) + '::' + ( post_data.excerpt || '' );
-	}
-
-	return ( $('#title').val() || '' ) + '::' + ( $('#content').val() || '' ) + '::' + ( $('#excerpt').val() || '' );
-};
-
 wp.autosave.local = {
-
 	lastSavedData: '',
 	blog_id: 0,
 	hasStorage: false,
@@ -404,9 +110,9 @@
 		var test = Math.random(), result = false;
 
 		try {
-			sessionStorage.setItem('wp-test', test);
-			result = sessionStorage.getItem('wp-test') == test;
-			sessionStorage.removeItem('wp-test');
+			sessionStorage.setItem( 'wp-test', test );
+			result = sessionStorage.getItem( 'wp-test' ) === test;
+			sessionStorage.removeItem( 'wp-test' );
 		} catch(e) {}
 
 		this.hasStorage = result;
@@ -424,10 +130,11 @@
 		if ( this.hasStorage && this.blog_id ) {
 			stored_obj = sessionStorage.getItem( 'wp-autosave-' + this.blog_id );
 
-			if ( stored_obj )
+			if ( stored_obj ) {
 				stored_obj = JSON.parse( stored_obj );
-			else
+			} else {
 				stored_obj = {};
+			}
 		}
 
 		return stored_obj;
@@ -460,8 +167,9 @@
 	getData: function() {
 		var stored = this.getStorage(), post_id = $('#post_ID').val();
 
-		if ( !stored || !post_id )
+		if ( !stored || !post_id ) {
 			return false;
+		}
 
 		return stored[ 'post_' + post_id ] || false;
 	},
@@ -477,15 +185,17 @@
 	setData: function( stored_data ) {
 		var stored = this.getStorage(), post_id = $('#post_ID').val();
 
-		if ( !stored || !post_id )
+		if ( !stored || !post_id ) {
 			return false;
+		}
 
-		if ( stored_data )
+		if ( stored_data ) {
 			stored[ 'post_' + post_id ] = stored_data;
-		else if ( stored.hasOwnProperty( 'post_' + post_id ) )
+		} else if ( stored.hasOwnProperty( 'post_' + post_id ) ) {
 			delete stored[ 'post_' + post_id ];
-		else
+		} else {
 			return false;
+		}
 
 		return this.setStorage(stored);
 	},
@@ -502,30 +212,31 @@
 	save: function( data ) {
 		var result = false, post_data, compareString;
 
-		if ( ! data ) {
-			post_data = wp.autosave.getPostData();
-		} else {
+		if ( data ) {
 			post_data = this.getData() || {};
 			$.extend( post_data, data );
-			post_data.autosave = true;
+		} else {
+			post_data = wp.autosave.getPostData('local');
+			// Cannot get the post data at the moment
+			if ( ! post_data ) {
+				return false;
+			}
 		}
 
-		// Cannot get the post data at the moment
-		if ( ! post_data.autosave )
-			return false;
-
 		compareString = wp.autosave.getCompareString( post_data );
 
 		// If the content, title and excerpt did not change since the last save, don't save again
-		if ( compareString == this.lastSavedData )
+		if ( compareString === this.lastSavedData ) {
 			return false;
+		}
 
-		post_data['save_time'] = (new Date()).getTime();
-		post_data['status'] = $('#post_status').val() || '';
+		post_data.save_time = ( new Date() ).getTime();
+		post_data.status = $( '#post_status' ).val() || '';
 		result = this.setData( post_data );
 
-		if ( result )
+		if ( result ) {
 			this.lastSavedData = compareString;
+		}
 
 		return result;
 	},
@@ -535,18 +246,22 @@
 		var self = this;
 
 		// Check if the browser supports sessionStorage and it's not disabled
-		if ( ! this.checkStorage() )
+		if ( ! this.checkStorage() ) {
 			return;
+		}
 
 		// Don't run if the post type supports neither 'editor' (textarea#content) nor 'excerpt'.
-		if ( ! $('#content').length && ! $('#excerpt').length )
+		if ( ! $('#content').length && ! $('#excerpt').length ) {
 			return;
+		}
 
-		if ( settings )
+		if ( settings ) {
 			$.extend( this, settings );
+		}
 
-		if ( !this.blog_id )
-			this.blog_id = typeof window.autosaveL10n != 'undefined' ? window.autosaveL10n.blog_id : 0;
+		if ( !this.blog_id ) {
+			this.blog_id = typeof window.autosaveL10n !== 'undefined' ? window.autosaveL10n.blog_id : 0;
+		}
 
 		$(document).ready( function(){ self.run(); } );
 	},
@@ -566,23 +281,24 @@
 			protect: true
 		});
 
-		$('form#post').on('submit.autosave-local', function() {
-			var editor = typeof tinymce != 'undefined' && tinymce.get('content'), post_id = $('#post_ID').val() || 0;
+		$( 'form#post' ).on( 'submit.autosave-local', function() {
+			var editor = typeof tinymce !== 'undefined' && tinymce.get('content'),
+				post_id = $('#post_ID').val() || 0;
 
 			if ( editor && ! editor.isHidden() ) {
 				// Last onSubmit event in the editor, needs to run after the content has been moved to the textarea.
 				editor.onSubmit.add( function() {
 					wp.autosave.local.save({
-						post_title: $('#title').val() || '',
-						content: $('#content').val() || '',
-						excerpt: $('#excerpt').val() || ''
+						post_title: $( '#title' ).val() || '',
+						content: $( '#content' ).val() || '',
+						excerpt: $( '#excerpt' ).val() || ''
 					});
 				});
 			} else {
 				self.save({
-					post_title: $('#title').val() || '',
-					content: $('#content').val() || '',
-					excerpt: $('#excerpt').val() || ''
+					post_title: $( '#title' ).val() || '',
+					content: $( '#content' ).val() || '',
+					excerpt: $( '#excerpt' ).val() || ''
 				});
 			}
 
@@ -592,11 +308,11 @@
 
 	// Strip whitespace and compare two strings
 	compare: function( str1, str2 ) {
-		function remove( string ) {
+		function removeSpaces( string ) {
 			return string.toString().replace(/[\x20\t\r\n\f]+/g, '');
 		}
 
-		return ( remove( str1 || '' ) == remove( str2 || '' ) );
+		return ( removeSpaces( str1 || '' ) === removeSpaces( str2 || '' ) );
 	},
 
 	/**
@@ -607,16 +323,20 @@
 	 * @return void
 	 */
 	checkPost: function() {
-		var self = this, post_data = this.getData(), content, post_title, excerpt, notice,
-			post_id = $('#post_ID').val() || 0, cookie = wpCookies.get( 'wp-saving-post-' + post_id );
+		var content, post_title, excerpt, notice,
+			self = this,
+			post_data = this.getData(),
+			post_id = $( '#post_ID' ).val() || 0,
+			cookie = wpCookies.get( 'wp-saving-post-' + post_id );
 
-		if ( ! post_data )
+		if ( ! post_data ) {
 			return;
+		}
 
 		if ( cookie ) {
 			wpCookies.remove( 'wp-saving-post-' + post_id );
 
-			if ( cookie == 'saved' ) {
+			if ( cookie === 'saved' ) {
 				// The post was saved properly, remove old data and bail
 				this.setData( false );
 				return;
@@ -624,18 +344,21 @@
 		}
 
 		// There is a newer autosave. Don't show two "restore" notices at the same time.
-		if ( $('#has-newer-autosave').length )
+		if ( $( '#has-newer-autosave' ).length ) {
 			return;
+		}
 
-		content = $('#content').val() || '';
-		post_title = $('#title').val() || '';
-		excerpt = $('#excerpt').val() || '';
+		content = $( '#content' ).val() || '';
+		post_title = $( '#title' ).val() || '';
+		excerpt = $( '#excerpt' ).val() || '';
 
-		if ( $('#wp-content-wrap').hasClass('tmce-active') && typeof switchEditors != 'undefined' )
+		if ( $( '#wp-content-wrap' ).hasClass( 'tmce-active' ) && typeof switchEditors !== 'undefined' ) {
 			content = switchEditors.pre_wpautop( content );
+		}
 
 		// cookie == 'check' means the post was not saved properly, always show #local-storage-notice
-		if ( cookie != 'check' && this.compare( content, post_data.content ) && this.compare( post_title, post_data.post_title ) && this.compare( excerpt, post_data.excerpt ) ) {
+		if ( cookie !== 'check' && this.compare( content, post_data.content ) &&
+				this.compare( post_title, post_data.post_title ) && this.compare( excerpt, post_data.excerpt ) ) {
 			return;
 		}
 
@@ -646,20 +369,20 @@
 			excerpt: excerpt
 		};
 
-		notice = $('#local-storage-notice');
-		$('.wrap h2').first().after( notice.addClass('updated').show() );
+		notice = $( '#local-storage-notice' );
+		$('.wrap h2').first().after( notice.addClass( 'updated' ).show() );
 
-		notice.on( 'click', function(e) {
+		notice.on( 'click.autosae-local', function(e) {
 			var target = $( e.target );
 
-			if ( target.hasClass('restore-backup') ) {
+			if ( target.hasClass( 'restore-backup' ) ) {
 				self.restorePost( self.restore_post_data );
 				target.parent().hide();
-				$(this).find('p.undo-restore').show();
-			} else if ( target.hasClass('undo-restore-backup') ) {
+				$(this).find( 'p.undo-restore' ).show();
+			} else if ( target.hasClass( 'undo-restore-backup' ) ) {
 				self.restorePost( self.undo_post_data );
 				target.parent().hide();
-				$(this).find('p.local-restore').show();
+				$(this).find( 'p.local-restore' ).show();
 			}
 
 			e.preventDefault();
@@ -674,20 +397,21 @@
 			// Set the last saved data
 			this.lastSavedData = wp.autosave.getCompareString( post_data );
 
-			if ( $('#title').val() != post_data.post_title )
-				$('#title').focus().val( post_data.post_title || '' );
+			if ( $( '#title' ).val() !== post_data.post_title ) {
+				$( '#title' ).focus().val( post_data.post_title || '' );
+			}
 
-			$('#excerpt').val( post_data.excerpt || '' );
-			editor = typeof tinymce != 'undefined' && tinymce.get('content');
+			$( '#excerpt' ).val( post_data.excerpt || '' );
+			editor = typeof tinymce !== 'undefined' && tinymce.get('content');
 
-			if ( editor && ! editor.isHidden() && typeof switchEditors != 'undefined' ) {
+			if ( editor && ! editor.isHidden() && typeof switchEditors !== 'undefined' ) {
 				// Make sure there's an undo level in the editor
 				editor.undoManager.add();
 				editor.setContent( post_data.content ? switchEditors.wpautop( post_data.content ) : '' );
 			} else {
 				// Make sure the Text editor is selected
-				$('#content-html').click();
-				$('#content').val( post_data.content );
+				$( '#content-html' ).click();
+				$( '#content' ).val( post_data.content );
 			}
 
 			return true;
@@ -697,6 +421,145 @@
 	}
 };
 
+// Autosave on the server
+wp.autosave.server = {
+	autosaveLast: '',
+
+	init: function() {
+		var self = this;
+
+		$document.on( 'heartbeat-send.autosave', function( event, data ) {
+			var autosaveData = self.getData();
+
+			if ( autosaveData ) {
+				data.wp_autosave = autosaveData;
+			}
+		}).on( 'heartbeat-tick.autosave', function( event, data ) {
+			if ( data.wp_autosave ) {
+				self.response( data.wp_autosave );
+			}
+		}).on( 'heartbeat-connection-lost.autosave', function( event, error, status ) {
+			// When connection is lost, keep user from submitting changes.
+			if ( 'timeout' === error || 603 === status ) {
+				var notice = $('#lost-connection-notice');
+
+				if ( ! wp.autosave.local.hasStorage ) {
+					notice.find('.hide-if-no-sessionstorage').hide();
+				}
+
+				notice.show();
+				wp.autosave.disableButtons();
+			}
+		}).on( 'heartbeat-connection-restored.autosave', function() {
+			$('#lost-connection-notice').hide();
+			wp.autosave.enableButtons();
+		}).ready( function() {
+			self._schedule();
+
+			if ( $('#wp-content-wrap').hasClass('tmce-active') && typeof switchEditors !== 'undefined' ) {
+				self.autosaveLast = wp.autosave.getCompareString({
+					post_title : $( '#title' ).val() || '',
+					content : switchEditors.pre_wpautop( $( '#content' ).val() ) || '',
+					excerpt : $( '#excerpt' ).val() || ''
+				});
+			} else {
+				self.autosaveLast = wp.autosave.getCompareString();
+			}
+
+			// Autosave new posts after a title is typed
+			if ( $( '#auto_draft' ).val() ) {
+				$( '#title' ).blur( function() {
+					if ( ! this.value || $( '#auto_draft' ).val() !== '1' ) {
+						return;
+					}
+
+					self.triggerSave();
+				});
+			}
+		});
+	},
+
+	// Block saving for the next 10 sec.
+	blockSave: function() {
+		_blockSave = true;
+		window.clearTimeout( _blockSaveTimer );
+		_blockSaveTimer = window.setTimeout( function() {
+			_blockSave = false;
+		}, 10000 );
+	},
+
+	// Runs on heartbeat-response
+	response: function( data ) {
+		this._schedule();
+		_blockSave = false;
+		this.autosaveLast = this.lastCompareString;
+		this.lastCompareString = '';
+
+		$document.trigger( 'after-autosave', [data] );
+		$( '.autosave-message' ).text( data.message );
+		wp.autosave.enableButtons();
+
+		if ( data.success ) {
+			// No longer an auto-draft
+			$( '#auto_draft' ).val('');
+		}
+	},
+
+	disable: function() {
+		_disabled = true;
+	},
+
+	_schedule: function() {
+		nextRun = ( new Date() ).getTime() + ( autosaveL10n.autosaveInterval * 1000 ) || 60000;
+	},
+
+	// Reset the timing and tell heartbeat to connect now
+	triggerSave: function() {
+		nextRun = 0;
+		wp.heartbeat.connectNow();
+	},
+
+	// Runs on 'heartbeat-send'
+	getData: function() {
+		var compareString, postData;
+
+		if ( _disabled || _blockSave ) {
+			return false;
+		}
+
+		if ( ( new Date() ).getTime() < nextRun ) {
+			return false;
+		}
+
+		postData = wp.autosave.getPostData();
+
+		// post_data.content cannot be retrieved at the moment?
+		if ( ! postData ) {
+			return false;
+		}
+
+		compareString = wp.autosave.getCompareString( postData );
+
+		// Nothing to save or no change.
+		if ( compareString === this.autosaveLast ) {
+			return false;
+		}
+
+		this.blockSave();
+		wp.autosave.disableButtons();
+		this.lastCompareString = compareString;
+
+		$document.trigger( 'wpcountwords', [ postData.content ] )
+			.trigger( 'before-autosave', [ postData ] );
+
+		$( '.autosave-message' ).text( autosaveL10n.savingText );
+		postData._wpnonce = $( '#_wpnonce' ).val() || '';
+
+		return postData;
+	}
+};
+
 wp.autosave.local.init();
+wp.autosave.server.init();
 
-}(jQuery));
+}( jQuery, window ));
Index: tests/phpunit/includes/testcase-ajax.php
===================================================================
--- tests/phpunit/includes/testcase-ajax.php	(revision 26185)
+++ tests/phpunit/includes/testcase-ajax.php	(working copy)
@@ -42,12 +42,12 @@
 		'oembed_cache', 'image-editor', 'delete-comment', 'delete-tag', 'delete-link',
 		'delete-meta', 'delete-post', 'trash-post', 'untrash-post', 'delete-page', 'dim-comment',
 		'add-link-category', 'add-tag', 'get-tagcloud', 'get-comments', 'replyto-comment',
-		'edit-comment', 'add-menu-item', 'add-meta', 'add-user', 'autosave', 'closed-postboxes',
+		'edit-comment', 'add-menu-item', 'add-meta', 'add-user', 'closed-postboxes',
 		'hidden-columns', 'update-welcome-panel', 'menu-get-metabox', 'wp-link-ajax',
 		'menu-locations-save', 'menu-quick-search', 'meta-box-order', 'get-permalink',
 		'sample-permalink', 'inline-save', 'inline-save-tax', 'find_posts', 'widgets-order',
 		'save-widget', 'set-post-thumbnail', 'date_format', 'time_format', 'wp-fullscreen-save-post',
-		'wp-remove-post-lock', 'dismiss-wp-pointer', 'nopriv_autosave'
+		'wp-remove-post-lock', 'dismiss-wp-pointer', 'heartbeat', 'nopriv_heartbeat',
 	);
 
 	/**
Index: tests/phpunit/tests/ajax/Autosave.php
===================================================================
--- tests/phpunit/tests/ajax/Autosave.php	(revision 26185)
+++ tests/phpunit/tests/ajax/Autosave.php	(working copy)
@@ -23,92 +23,158 @@
 	protected $_post = null;
 
 	/**
+	 * user_id
+	 * @var int
+	 */
+	protected $user_id = 0;
+
+	/**
 	 * Set up the test fixture
 	 */
 	public function setUp() {
 		parent::setUp();
+		// Set a user so the $post has 'post_author'
+		$this->user_id = $this->factory->user->create( array( 'role' => 'administrator' ) );
+		wp_set_current_user( $this->user_id );
+		
 		$post_id = $this->factory->post->create( array( 'post_status' => 'draft' ) );
 		$this->_post = get_post( $post_id );
 	}
 
 	/**
+	 * Tear down the test fixture.
+	 * Reset the current user
+	 */
+	public function tearDown() {
+		parent::tearDown();
+		wp_set_current_user( 0 );
+	}
+
+	/**
 	 * Test autosaving a post
 	 * @return void
 	 */
 	public function test_autosave_post() {
+		// The original post_author
+		wp_set_current_user( $this->user_id );
 
-		// Become an admin
-		$this->_setRole( 'administrator' );
-
 		// Set up the $_POST request
 		$md5 = md5( uniqid() );
 		$_POST = array(
-		    'post_id'       => $this->_post->ID,
-		    'autosavenonce' => wp_create_nonce( 'autosave' ),
-		    'post_content'  => $this->_post->post_content . PHP_EOL . $md5,
-			'post_type'     => 'post',
-		    'autosave'      => 1,
+			'action' =>	'heartbeat',
+			'_nonce' => wp_create_nonce( 'heartbeat-nonce' ),
+			'data' => array(
+				'wp_autosave' => array(
+				    'post_id'       => $this->_post->ID,
+				    '_wpnonce'      => wp_create_nonce( 'update-post_' . $this->_post->ID ),
+				    'post_content'  => $this->_post->post_content . PHP_EOL . $md5,
+					'post_type'     => 'post',
+				),
+			),
 		);
 
 		// Make the request
 		try {
-			$this->_handleAjax( 'autosave' );
+			$this->_handleAjax( 'heartbeat' );
 		} catch ( WPAjaxDieContinueException $e ) {
 			unset( $e );
 		}
 
-		// Get the response
-		$xml = simplexml_load_string( $this->_last_response, 'SimpleXMLElement', LIBXML_NOCDATA );
+		// Get the response, it is in heartbeat's response
+		$response = json_decode( $this->_last_response, true );
 
 		// Ensure everything is correct
-		$this->assertEquals( $this->_post->ID, (int) $xml->response[0]->autosave['id'] );
-		$this->assertEquals( 'autosave_' . $this->_post->ID, (string) $xml->response['action']);
+		$this->assertNotEmpty( $response['wp_autosave'] );
+		$this->assertTrue( $response['wp_autosave']['success'] );
 
 		// Check that the edit happened
-		$post = get_post( $this->_post->ID) ;
+		$post = get_post( $this->_post->ID );
 		$this->assertGreaterThanOrEqual( 0, strpos( $post->post_content, $md5 ) );
 	}
-
+	
 	/**
-	 * Test with an invalid nonce
+	 * Test autosaving a locked post
 	 * @return void
 	 */
-	public function test_with_invalid_nonce( ) {
+	public function test_autosave_locked_post() {
+		// Lock the post to another user
+		$another_user_id = $this->factory->user->create( array( 'role' => 'editor' ) );
+		wp_set_current_user( $another_user_id );
+		wp_set_post_lock( $this->_post->ID );
+		
+		wp_set_current_user( $this->user_id );
 
-		// Become an administrator
-		$this->_setRole( 'administrator' );
+		// Ensure post is locked
+		$this->assertEquals( $another_user_id, wp_check_post_lock( $this->_post->ID ) );
 
 		// Set up the $_POST request
+		$md5 = md5( uniqid() );
 		$_POST = array(
-		    'post_id'       => $this->_post->ID,
-		    'autosavenonce' => md5( uniqid() ),
-		    'autosave'      => 1
+			'action' =>	'heartbeat',
+			'_nonce' => wp_create_nonce( 'heartbeat-nonce' ),
+			'data' => array(
+				'wp_autosave' => array(
+				    'post_id'       => $this->_post->ID,
+				    '_wpnonce'      => wp_create_nonce( 'update-post_' . $this->_post->ID ),
+				    'post_content'  => $this->_post->post_content . PHP_EOL . $md5,
+					'post_type'     => 'post',
+				),
+			),
 		);
 
 		// Make the request
-		$this->setExpectedException( 'WPAjaxDieStopException', '-1' );
-		$this->_handleAjax( 'autosave' );
+		try {
+			$this->_handleAjax( 'heartbeat' );
+		} catch ( WPAjaxDieContinueException $e ) {
+			unset( $e );
+		}
+
+		$response = json_decode( $this->_last_response, true );
+
+		// Ensure everything is correct
+		$this->assertNotEmpty( $response['wp_autosave'] );
+		$this->assertTrue( $response['wp_autosave']['success'] );
+
+		// Check that the original post was NOT edited
+		$post = get_post( $this->_post->ID );
+		$this->assertFalse( strpos( $post->post_content, $md5 ) );
+
+		// Check if the autosave post was created
+		$autosave = wp_get_post_autosave( $this->_post->ID, get_current_user_id() );
+		$this->assertNotEmpty( $autosave );
+		$this->assertGreaterThanOrEqual( 0, strpos( $autosave->post_content, $md5 ) );
 	}
 
 	/**
-	 * Test with a bad post id
+	 * Test with an invalid nonce
 	 * @return void
 	 */
-	public function test_with_invalid_post_id( ) {
+	public function test_with_invalid_nonce( ) {
 
-		// Become an administrator
-		$this->_setRole( 'administrator' );
+		wp_set_current_user( $this->user_id );
 
 		// Set up the $_POST request
 		$_POST = array(
-		    'post_id'       => 0,
-		    'autosavenonce' => wp_create_nonce( 'autosave' ),
-		    'autosave'      => 1,
-			'post_type'     => 'post'
+			'action' =>	'heartbeat',
+			'_nonce' => wp_create_nonce( 'heartbeat-nonce' ),
+			'data' => array(
+				'wp_autosave' => array(
+				    'post_id'  => $this->_post->ID,
+				    '_wpnonce' => substr( md5( uniqid() ), 0, 10 ),
+				),
+			),
 		);
 
 		// Make the request
-		$this->setExpectedException( 'WPAjaxDieStopException', 'You are not allowed to edit this post.' );
-		$this->_handleAjax( 'autosave' );
+		try {
+			$this->_handleAjax( 'heartbeat' );
+		} catch ( WPAjaxDieContinueException $e ) {
+			unset( $e );
+		}
+
+		$response = json_decode( $this->_last_response, true );
+
+		$this->assertNotEmpty( $response['wp_autosave'] );
+		$this->assertFalse( $response['wp_autosave']['success'] );
 	}
 }

Index: src/wp-includes/revision.php
===================================================================
--- src/wp-includes/revision.php	(revision 26185)
+++ src/wp-includes/revision.php	(working copy)
@@ -215,7 +215,7 @@
  *
  * @param int|object|array $post Post ID, post object OR post array.
  * @param bool $autosave Optional. Is the revision an autosave?
  * @return mixed Null or 0 if error, new revision ID if success.
  */
 function _wp_put_post_revision( $post = null, $autosave = false ) {
 	if ( is_object($post) )
@@ -223,8 +223,8 @@
 	elseif ( !is_array($post) )
 		$post = get_post($post, ARRAY_A);
 
-	if ( !$post || empty($post['ID']) )
-		return;
+	if ( ! $post || empty($post['ID']) )
+		return new WP_Error( 'invalid_post', __( 'Invalid post ID' ) );
 
 	if ( isset($post['post_type']) && 'revision' == $post['post_type'] )
 		return new WP_Error( 'post_type', __( 'Cannot create a revision of a revision' ) );
@@ -215,7 +215,7 @@
  * @param int|object|array $post Post ID, post object OR post array.
  * @param bool $autosave Optional. Is the revision an autosave?
- * @return mixed Null or 0 if error, new revision ID if success.
+ * @return mixed WP_Error or 0 if error, new revision ID if success.
  */
 function _wp_put_post_revision( $post = null, $autosave = false ) {